-- Time-stamp: <Sat Jul 26 1997 23:26:24 Stardate: [-31]9694.67 hwloidl>
--
-- Alpha-Beta Search
-- Board manipulation functions
--
-- Encoding of board:     Start position:
--                10               .  
--               /|\	          /|\  
--              7-8-9	         x-x-x 
--              |\|/|	         |\|/| 
--              4-5-6	         .-.-. 
--              |/|\|	         |/|\| 
--              1-2-3	         .-o-.
--               \|/	          \|/  
--                0   	           .   
-- Static evaluation function: 
--  Compute distance of my own position from a known good position.
--  In this case a good fct might be as simple as the number of possible next 
--  moves. For O perhaps weight by index (being on field 10 means victory).
-----------------------------------------------------------------------------

--@node Board Manipulation, ,Top, Top
--@chapter Board Manipulation

module Board where

--@menu
--* Imports::			
--* Types::			
--* Show Fcts::			
--* Update and Insert::		
--* Check::			
--* Inital Board::		
--* Instances::			
--@end menu

--@node Imports, Types, Board Manipulation, Board Manipulation
--@section Imports

--@include Wins.hs
import Wins

#if defined(STRATEGIES)
--@ifset strategies
--@include Strategies.lhs
--@end ifset
import Strategies
#else
--@ifclear strategies
--@include ParForce.hs
--@end ifclear
-- import ParForce
{- no imports! -}
#endif

--@node Types, Show Fcts, Imports, Board Manipulation
--@section Types

type Board = [Piece]  -- position encoding see pic above 
#if (__HASKELL1__==4)
data Piece = X | O | Empty deriving (Show,Eq)
                                    {-!NFData!-}
#else
data Piece = X | O | Empty deriving (Text,Eq)
                                    {-!NFData!-}
#endif
				    
--@node Show Fcts, Update and Insert, Types, Board Manipulation
--@section Show Fcts

showBoard :: Board -> String
showBoard = show . (map showPiece)
--showBoard xs = "   " ++ (showPiece (xs!!10)) ++ "   \n" ++
--               "  /|\\  \n" ++
--	       " " ++ (showPiece (xs!!7)) ++ "-" ++ (showPiece (xs!!8)) ++ "-" ++ (showPiece (xs!!9)) ++ " \n" ++
--               " |\\|/| \n" ++
--	       " " ++ (showPiece (xs!!4)) ++ "-" ++ (showPiece (xs!!5)) ++ "-" ++ (showPiece (xs!!6)) ++ " \n" ++
--               " |/|\\| \n" ++
--	       " " ++ (showPiece (xs!!1)) ++ "-" ++ (showPiece (xs!!2)) ++ "-" ++ (showPiece (xs!!3)) ++ " \n" ++
--	       "  \\|/  \n" ++
--	       "   " ++ (showPiece (xs!!10)) ++ "   \n"

showPiece :: Piece -> String
showPiece X = "X"
showPiece O = "O"
showPiece Empty = "."

possb_moves :: Piece -> [[Int]]
possb_moves O = [{-0-}[1,2,3], {-1-}[0,1,2,4,5], {-2-}[0,1,3,5], {-3-}[0,2,5,6],
                 {-4-}[1,5,7], {-5-}[1,2,3,4,6,7,8], {-6-}[3,5,9], 
		 {-7-}[4,5,8,10], {-8-}[5,7,9,10], {-9-}[5,6,8,10], {-10-}[7,8,9] ]
possb_moves X = [{-0-}[], {-1-}[0,2], {-2-}[0,1,3], {-3-}[0,2],
                 {-4-}[1,5], {-5-}[1,2,3,4,6], {-6-}[3,5], 
		 {-7-}[4,5,8], {-8-}[5,7,9], {-9-}[5,6,8], {-10-}[]]
possb_moves Empty = take 11 (repeat [])

--@node Update and Insert, Check, Show Fcts, Board Manipulation
--@section Update and Insert

newPositions :: Piece -> Board -> [Board]
#if defined(SIMPLE)
newPositions piece board = concat (map (movePiece piece board) 
				    [n | (p,n) <- zip board [0..10], 
				         piece == p ])
#else
newPositions piece@X board = concat (map (movePiece piece board) 
				    [n | (p,n) <- zip board [0..10], 
				         piece == p ])
newPositions piece@O board = concat (map (movePiece piece board) 
                                    -- look at high fields first!!
				    [n | (p,n) <- reverse (zip board [0..10]), 
				         piece == p ])
#endif				    

movePiece :: Piece -> Board -> Int -> [Board]
movePiece piece board n = [ -- move piece from n to n'
                            insertPiece piece (insertPiece Empty board n) n' 
                          | n' <- (possb_moves piece)!!n,
			    board!!n' == Empty]

insertPiece :: Piece -> Board -> Int -> Board
insertPiece piece board n 
 | piece==Empty && (board!!n) == Empty  = 
          error ("Can't insert empty on empty field" ++ (show n) ++ "\n")
 | piece /= Empty && (board!!n) /= Empty = 
          error ("Can't move piece " ++ (showPiece piece) ++ 
                 "to non-empty field " ++ (show n) ++ "\n")
 | otherwise           = take n board ++ [piece] ++ drop (n+1) board

--@node Check, Inital Board, Update and Insert, Board Manipulation
--@section Check

fullBoard O b = b!!10 == O                              -- escaped
fullBoard X b = and [ and [ notEmpty x
	                  | x <- (possb_moves p)!!n ]
	            | (p,n) <- zip b [0..10], p == X ] -- all fields
		where notEmpty x = not (b!!x == Empty)
             				 
--@node Inital Board, Instances, Check, Board Manipulation
--@section Inital Board

initialBoard :: Board
initialBoard = [Empty, Empty, O, Empty, Empty, Empty, Empty, X, X, X, Empty] 

#if (__HASKELL1__==4)
data Evaluation = XWin | OWin | Score Int deriving (Show,Eq)
                                                   {-!NFData!-}
#else
data Evaluation = XWin | OWin | Score Int deriving (Text,Eq)
                                                   {-!NFData!-}
#endif

{- OLD: partain
instance Eq Evaluation where
    XWin       == XWin	     = True
    OWin       == OWin	     = True
    (Score i1) == (Score i2) = i1 == i2
    _	       == _          = False
    a	  /= b	   = not (a == b)

instance Text Evaluation where
    showsPrec d XWin = showString "XWin"
    showsPrec d OWin = showString "OWin"
    showsPrec d (Score i) = showParen (d >= 10) showStr
	where
	  showStr = showString "Score" . showChar ' ' . showsPrec 10 i

    readsPrec p = error "no readsPrec for Evaluations"
    readList = error "no readList for Evaluations"
    showList []	= showString "[]"
    showList (x:xs)
		= showChar '[' . shows x . showl xs
		  where showl []     = showChar ']'
			showl (x:xs) = showChar ',' . shows x . showl xs
-}

instance Ord Evaluation where
 _ <= XWin    = True
 XWin <= _    = False
 OWin <= _    = True
 _ <= OWin    = False
 (Score x) <= (Score y) = x <= y

eval 100 = XWin
eval (-100) = OWin
eval x = Score x

static :: Piece -> Board -> Evaluation
#if defined(PAR_STATIC) && defined(STRATEGIES)
static X board = interpret 0 (parMap rnf (score board) wins)
static O board = (eval . negate . sum) $|| parList rnf $ 
                 [ length ((possb_moves O)!!n) + 2*n 
                 | (p,n) <- zip board [0..10], p==O]
#else
static X board = interpret 0 (map (score board) wins)
static O board = (eval . negate . sum)  
                 [ length ((possb_moves O)!!n) + 2*n 
                 | (p,n) <- zip board [0..10], p==O]
#endif

interpret :: Int -> [Evaluation] -> Evaluation
interpret x [] = (Score x)
interpret x (Score y:l) = interpret (x+y) l
interpret x (XWin:l) = XWin
interpret x (OWin:l) = OWin

score :: Board -> Win -> Evaluation
score board win = eval (sum (zipWith scorePiece board win)) 
-- eval (sum (map sum (zipWith (zipWith scorePiece) board win)))

scorePiece :: Piece -> Int -> Int
scorePiece X score | score > 0 = score  -- my piece on a good position?
                   | otherwise = 0
scorePiece Empty _ = 0
scorePiece O score | score < 0 = -score  -- enemy piece on a bad position?
                   | otherwise = 0

{-
#if 0
-- This looks very much like a zipWith f to me
map2 :: (a -> b -> c) -> [a] -> [b] -> [c]
map2 f [] x = []
map2 f x [] = []
map2 f (x:xs) (y:ys) = f x y:map2 f xs ys
#endif
-}

--@node Instances,  , Inital Board, Board Manipulation
--@section Instances

#if !defined(HASKELL_13)
{-* DO NOT EDIT BELOW THIS POINT *-}
instance NFData Piece where{rnf (X)= ();rnf (O)= ();rnf (Empty)= ()}
instance NFData Evaluation where{rnf (XWin)= ();rnf (OWin)= ();rnf (Score a)=rnf a}
#endif